// 
// Written by Ryan Geiss, 2009
// http://www.geisswerks.com/
// 
// Converts "pfs First Choice" word processing (*not database*!) 
//   documents to HTML, for future generations.  :)
//
// Also copies the timestamp from your old file, to the new one.
//  
// Note that these documents usually had the extension ".doc".
//
// This source file was compiled with Visual Studio 2003, for x86.
// 

#include "stdafx.h"
#include <stdio.h>
#include <stdlib.h>
#include <windows.h>


int main(int argc, char* argv[])
{
    printf("\n");

    if (argc==1)
    {
        printf("USAGE: specify the filename (usually ending in .doc) to convert to HTML.\n");
        printf("\n");
        exit(1);
    }

    FILE* f = fopen(argv[1], "rb");
    if (!f)
    {
        printf("Error opening input file: %s\n", argv[1]);
        exit(2);
    }

    const char* infilename = argv[1];

    char outfilename[MAX_PATH];
    strcpy(outfilename, infilename);
    char *p = strrchr(outfilename, '.');
    if (p)
      strcpy(p, ".html");
    else
      strcat(outfilename, ".html");
    
    FILE* fout = fopen(outfilename, "wb");
    if (!fout)
    {
        printf("Error opening output file: %s\n", outfilename);
        exit(2);
    }

    const char *pShortOutFilename = strrchr(outfilename, '\\');
    if (pShortOutFilename)
      pShortOutFilename++;
    else
      pShortOutFilename = outfilename;

    printf("Converting pfs First Choice text document:\n");
    printf("    %s\n", infilename);
    printf("to HTML document:\n");
    printf("    %s\n", pShortOutFilename);
    printf("\n");

    fprintf(fout, "<HTML>\n");
    fprintf(fout, "<HEAD><TITLE>%s</TITLE></HEAD>\n", pShortOutFilename );
    fprintf(fout, "<BODY>\n");
    fprintf(fout, "<PRE>\n");
    fprintf(fout, "\n");



    // strip off the first 512 bytes:
    unsigned char buf[1024];
    fread(buf, 1, 512, f);
    
    unsigned char outbuf[256];
    int outchars = 0;

    bool bWasBold = false;
    bool bWasItalics = false;
    bool bWasUnderline = false;


    /*
    // In case you need to find some 4-digit codes, enable this:
    //for (int code=9400; code<9700; code++)
    for (int code=128; code<255; code++)
        fprintf(fout, "%04d = &#%04d;<BR>", code, code);
    /**/

    // process the rest:
    while (!feof(f))
    {
        int ch = fgetc(f);
        if (ch==-1)
          break;
    
        outbuf[outchars++] = (char)ch;

        // TRANSLATE:
        {
            /*
	          linefeeds:
		          0DE081 -> 0D0A  paragraph
		          0D80   -> 0D0A  newline
	          ascii chars:
		          [0xB0-0xF0] 0x88 -> just lose the 0x88
	          underlined/bold/both:   let 0xZZ == 0x81, 0x82, or 0x84
		          lowercase:  0xE1+ 0xZZ -> 0x61+    (lose the 0xZZ and subtract 0x80 from the other byte.)
		          uppercase:  0xC1+ 0xZZ -> 0x41+
            */

            bool bBold      = false;
            bool bItalics   = false;
            bool bUnderline = false;

            unsigned char *p = &outbuf[outchars-1]; // pointer to last char in string
            if (outchars >= 3 && *(p-2)==0x0D && *(p-1)==0xE0 && *p==0x81)
            {
              *(p-1) = 0x0A;
              #if 0
                outchars--;
                p--;
              #else
                const int num_spaces_to_indent = 0;
                for (int i=0; i<num_spaces_to_indent; i++)
                  p[i] = 0x20;
                outchars += num_spaces_to_indent-1;
                p += num_spaces_to_indent-1;
              #endif
            }
            else if (outchars >= 2 && *(p-1)==0x0D && *p==0x80)
            {
              *p = 0x0A;
            }
            else if (outchars >= 2 && (p[0]>=0x80 && p[0]<=0x8F) && p[-1]>=0x80)
            {
              const unsigned char ctrl = p[0];    // 1/2/4 bits = bold/italic/underline; 8 bit = extended ASCII char
              const unsigned char ch   = p[-1];

              // erase the code for now; we'll write the char out in a minute.
              // p always points to the LAST valid character in the stream.
              p -= 2;
              outchars -= 2;

              if (ctrl & 1)
              {
                if (!bWasUnderline)
                {
                  strcpy( (char*)(p+1), "<U>" );
                  outchars += 3;
                  p        += 3;
                }
                bUnderline = true;
              }
              if (ctrl & 2)
              {
                if (!bWasBold)
                {
                  strcpy( (char*)(p+1), "<B>" );
                  outchars += 3;
                  p        += 3;
                }
                bBold = true;
              }
              if (ctrl & 4)
              {
                if (!bWasItalics)
                {
                  strcpy( (char*)(p+1), "<EM>" );
                  outchars += 4;
                  p        += 4;
                }
                bItalics = true;
              }

              if (ctrl & 8)
              {
                // translate arbitrary ASCII code:

                int code = 0;
                switch(ch)
                {
                // translate codes for single-line box-drawing chars:
                case 0xB3: code = 9474; break;
                case 0xC4: code = 9472; break;
                case 0xDA: code = 9484; break;
                case 0xBF: code = 9488; break;
                case 0xC0: code = 9492; break;
                case 0xD9: code = 9496; break;
                case 0xB4: code = 9508; break;
                case 0xC3: code = 9500; break;
                case 0xC2: code = 9516; break;
                case 0xC5: code = 9532; break;
                case 0xC1: code = 9524; break;
                // accents:
                case 130:  code = 233; break; // e with accent to right
                case 133:  code = 224; break; // a with accent to *LEFT*
                case 138:  code = 232; break; // e with accent to *LEFT*
                case 144:  code = 201; break; // E with accent to right
                case 160:  code = 225; break; // a with accent to right
                case 161:  code = 237; break; // i with accent to right
                case 162:  code = 243; break; // o with accent to right
                case 163:  code = 250; break; // u with accent to right
                case 164:  code = 241; break; // n with tilde
                case 165:  code = 209; break; // N with tilde
                case 168:  code = 191; break; // upside down ?
                case 171:  code = 189; break; // 1/2
                case 172:  code = 188; break; // 1/4
                case 173:  code = 161; break; // upside down !
                case 174:  code = 171; break; // <<
                case 175:  code = 187; break; // >>
                case 241:  code = 177; break; // +-
                case 246:  code = 247; break; // divide
                //case 251:  code =    ; break; // sqrt
                case 253:  code = 178; break; // squared
                }

                int num_chars = 0;
                if (code)
                  num_chars = sprintf((char*)&p[1], "&#%04d;", code);
                else 
                  num_chars = sprintf((char*)&p[1], "&#%d;", ch);
                
                outchars += num_chars;
                p += num_chars;
              }
              else
              {
                // echo normal char
                p[1] = ch - 0x80;
                outchars++;
                p++;
              }


            }
        
            const unsigned char ch = *p;

            if (ch < 0x80)
            {
              if (!bUnderline && bWasUnderline)
              {
                strcpy( (char*)p, "</U>" );
                p[4] = ch;
                outchars += 4;
                p += 4;
                bUnderline = false;
              }
              if (!bBold && bWasBold)
              {
                strcpy( (char*)p, "</B>" );
                p[4] = ch;
                outchars += 4;
                p += 4;
                bBold = false;
              }
              if (!bItalics && bWasItalics)
              {
                strcpy( (char*)p, "</EM>" );
                p[5] = ch;
                outchars += 5;
                p += 5;
                bItalics = false;
              }

              bWasBold      = bBold;
              bWasItalics   = bItalics;
              bWasUnderline = bUnderline;
            }
        }

        if (outchars > 240)
        {
          fwrite(outbuf, 1, outchars, fout);
          outchars = 0;
        }
    }

    fprintf(fout, "\n");
    fprintf(fout, "\n");
    fprintf(fout, "</PRE>\n");
    fprintf(fout, "</BODY>\n");
    fprintf(fout, "</HTML>\n");

    fclose(f);
    fclose(fout);

    // ------ fix dest file time ---------

    HANDLE hFile = CreateFile(infilename, FILE_READ_ATTRIBUTES, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, 0);
    if (hFile != INVALID_HANDLE_VALUE)
    {
      FILETIME x,y,z;
      if (GetFileTime(hFile, &x, &y, &z))
      {
        HANDLE hFile2 = CreateFile(outfilename, FILE_WRITE_ATTRIBUTES, FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, 0);
        if (hFile2 != INVALID_HANDLE_VALUE)
        {
          SetFileTime(hFile2, &x, &y, &z);
          CloseHandle(hFile2);
        }
      }        
      CloseHandle(hFile);
    }


    printf("Done!\n");
    printf("\n");

}

